function [v origin camcosy] = projectpoints(data, campos, varargin)
% [v origin camcosy] = projectpoints(data, cammat, campos [,varargin])
% Project 2D image points in the n-by-2 matrix data onto vectors from
% origin (the camera's optical center) with the directions stored in the
% n-by-3 matrix v.
% campos : The camera position [x y z phiz] in world coordinates [mm, �].
%          This should be for the camera in centered position, meaning that in
%	   our setups the camera may only be rotated around the z axis.
%          The rotations because of pan and tilt are calculated independently.
% varargin:
% 'pan'    : Camera pan value (ticks)
% 'tilt'   : Camera tilt value (ticks)
% 'imscale': Image scaling factor
% 'ppoint' : Principal point of camera
% 'dcoeff' : Coefficients [k1 k2] of radial lens distortion
% 'cammat' : Camera projection matrix (intrinsic parameters)

  % Camera parameters for Sony EVI D70P (as provided in the manual)
  % TODO: Read these from a config file.
  MINPANVALUE = -2277;		% Minimum pan motor tick value
  MAXPANVALUE = 2277;		% Maximum pan motor tick value
  MINPANANGLE = -170;		% Minimum pan angle from center position
  MAXPANANGLE = 170;		% Maximum pan angle from center position
  MINTILTVALUE = -1199;		% Minimum tilt motor tick value
  MAXTILTVALUE = 407;		% Maximum tilt motor tick value
  MINTILTANGLE = -90;		% Minimum tilt angle from center position
  MAXTILTANGLE = 30;		% Maximum tilt angle from center position
  RESOLUTION = [795 596];	% physical resolution of sensor
  PIXELOFFSET = [3 12];		% blind sensor elements
  FOCALLENGTH = 4.1;		% Focal length in mm (assuming zoom 0)
  VIEWANGLE = 48;		% lens viewing angle (assuming zoom 0)
  CELLSIZE = [0.00485 0.00465];	% physical size of image elements
  
  pan = 0;
  tilt = 0;
  imscale = 1.0;
  ppoint = [];
  v = [];
  origin = [];
  dcoeff = [];
  T = [];
  
  if nargin > 3
    for i=1:length(varargin)
      if strcmp(varargin{i}, 'pan')
	pan = varargin{i+1};
      elseif strcmp(varargin{i}, 'tilt')
	tilt = varargin{i+1};
      elseif strcmp(varargin{i}, 'imscale')
	imscale = varargin{i+1};
      elseif strcmp(varargin{i}, 'ppoint')
	ppoint = varargin{i+1};
      elseif strcmp(varargin{i}, 'cammat')
	T = varargin{i+1};
      elseif strcmp(varargin{i}, 'dcoeff')
	dcoeff = varargin{i+1};
      end
    end
  end
  
  if length(ppoint) == 0
    ppoint = floor((RESOLUTION/2-PIXELOFFSET)*imscale);
  end
  
  % If parameters are given, compensate for radial distortion
  %if length(dcoeff) > 0
  %  rsquare = (data(:,1)-ppoint(1)).^2 + (data(:,2)-ppoint(2)).^2;
    
  %  data(:,1) = data(:,1)+(data(:,1)-ppoint(1)).*(dcoeff(1)*rsquare+dcoeff(2)*rsquare.*rsquare);
  %  data(:,2) = data(:,2)+(data(:,2)-ppoint(2)).*(dcoeff(2)*rsquare+dcoeff(2)*rsquare.*rsquare);
  %  data(:,1) = (1.0/(1+dcoeff(1)*rsquare+dcoeff(2)*rsquare.*rsquare))'.*data(:,1);
  %  data(:,2) = (1.0/(1+dcoeff(1)*rsquare+dcoeff(2)*rsquare.*rsquare))'.*data(:,2);
  %end

  % Calculate transformation matrix between global and camera coordinates
  % This is accomplished by calculating the rotation angles for camera position,
  % camera pan and tilt, and then forming a combined transformation matrix.
  tiltangle = ((MAXTILTANGLE-MINTILTANGLE)/(MAXTILTVALUE-MINTILTVALUE)*(tilt))*pi/180;
  % Panning sign is inverted because of overhead camera mounting
  panangle = ((MAXPANANGLE-MINPANANGLE)/(MAXPANVALUE-MINPANVALUE)*(-pan)-campos(4))*pi/180;

  % pan is a rotation around the z axis
  panrot = [cos(panangle) -sin(panangle) 0; sin(panangle) cos(panangle) 0; 0 0 1];
  % tilt is a rotation around the x axis
  tiltrot = [1 0 0; 0 cos(tiltangle) -sin(tiltangle); 0 sin(tiltangle) cos(tiltangle)];
  rot = panrot*tiltrot;
  
  % The intrinsic camera transformation matrix (if not provided)
  % We assume that the image axes are perpendicular
  if length(T) == 0
    T = [FOCALLENGTH*imscale/CELLSIZE(1) 0 ppoint(1) 0; 
	0 FOCALLENGTH*imscale/CELLSIZE(2) ppoint(2) 0;
	0 0                               1 0];
  end

  % The origin is given (approximately) by the camera position, assuming
  % that it corresponds to the camera projection center.
  origin = campos(1:3);
  
  % The line corresponding to a point on the image plane is given by the
  % inverse transformation with T. This transforms pixel positions
  % to 3D coordinates in the camera coordinate system.
  %v = (T\data')
  %v = v(1:3,:)

  v(1,:) = (data(:,1)-T(1,3))/T(1,1);
  v(2,:) = (data(:,2)-T(2,3))/T(2,2);
  v(3,:) = 1;

  % The camera coordinate system in base position is formed by rotating 90 degree
  % counter-clockwise around the x-axis. Hence, the point position also have to be
  % rotated this way.
  camcosy = [1 0 0; 0 0 1; 0 -1 0];
  
  % Rotate camera coordinate system into world coordinate system
  camcosy = rot*camcosy;
  
  % Project point coordinates into (rotated) camera coordinate system
  v = camcosy*v;
  
  % Finally, normalize vectors to unit length 
  for i=1:size(v,2)
    v(:,i) = v(:,i)/norm(v(:,i));
  end

  %linestyles = {'g.-' 'b.-' 'c.-' 'm.-' 'k.-'};
  %cnt = 0;
  %for i=1:size(v,2)
  %    plot3(campos(1), campos(2), campos(3), 'r*');
  %    plot3([campos(1) campos(1)+5000*v(1,i)], [campos(2) campos(2)+5000*v(2,i)], [campos(3) campos(3)+5000*v(3,i)], linestyles{mod(cnt, length(linestyles))+1});
  %    xlabel('X');
  %    ylabel('Y');
  %    zlabel('Z');
  %    cnt = cnt+1;
  %end

end